"""parses [PREDICT] section of config"""
import os
from pathlib import Path

import attr
from attr import converters, validators
from attr.validators import instance_of

from .converters import comma_separated_list, expanded_user_path
from .validators import is_a_directory, is_a_file, is_valid_model_name
from .. import device


@attr.s
class PredictConfig:
    """class that represents [PREDICT] section of config.toml file

    Attributes
    ----------
    csv_path : str
        path to where dataset was saved as a csv.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    labelmap_path : str
        path to 'labelmap.json' file.
    models : list
        of model names. e.g., 'models = TweetyNet, GRUNet, ConvNet'
    batch_size : int
        number of samples per batch presented to models during training.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    device : str
        Device on which to work with model + data.
        Defaults to 'cuda' if torch.cuda.is_available is True.
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    annot_csv_filename : str
        name of .csv file containing predicted annotations.
        Default is None, in which case the name of the dataset .csv
        is used, with '.annot.csv' appended to it.
    output_dir : str
        path to location where .csv containing predicted annotation
        should be saved. Defaults to current working directory.
    min_segment_dur : float
        minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from lbl_tb. Default is None, in which case no
        segments are removed.
    majority_vote : bool
        if True, transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment. This transform can only be
        applied if the labelmap contains an 'unlabeled' label,
        because unlabeled segments makes it possible to identify
        the labeled segments. Default is False.
    """
    # required, external files
    checkpoint_path = attr.ib(converter=expanded_user_path,
                              validator=is_a_file)
    labelmap_path = attr.ib(converter=expanded_user_path,
                            validator=is_a_file)

    # required, model / dataloader
    models = attr.ib(converter=comma_separated_list,
                     validator=[instance_of(list), is_valid_model_name])
    batch_size = attr.ib(converter=int, validator=instance_of(int))

    # csv_path is actually 'required' but we can't enforce that here because cli.prep looks at
    # what sections are defined to figure out where to add csv_path after it creates the csv
    csv_path = attr.ib(converter=converters.optional(expanded_user_path),
                       validator=validators.optional(is_a_file),
                       default=None
                       )

    # optional, transform
    spect_scaler_path = attr.ib(converter=converters.optional(expanded_user_path),
                                validator=validators.optional(is_a_file),
                                default=None)

    # optional, data loader
    num_workers = attr.ib(validator=instance_of(int), default=2)
    device = attr.ib(validator=instance_of(str), default=device.get_default())

    annot_csv_filename = attr.ib(validator=validators.optional(instance_of(str)), default=None)
    output_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory, default=Path(os.getcwd()))
    min_segment_dur = attr.ib(validator=validators.optional(instance_of(float)), default=None)
    majority_vote = attr.ib(validator=instance_of(bool), default=True)


REQUIRED_PREDICT_OPTIONS = [
    'checkpoint_path',
    'labelmap_path',
    'models',
]


def parse_predict_config(config_obj, config_path):
    """parse [PREDICT] section of config.toml file

    Parameters
    ----------
    config_obj : ConfigParser
        containing config.toml file already loaded by parse function
    config_path : str
        path to config.toml file (used for error messages)

    Returns
    -------
    predict_config : vak.config.predict.PredictConfig
        instance of PredictConfig class that represents [PREDICT] section
        of config.toml file
    """
    predict_section = dict(
        config_obj['PREDICT'].items()
    )

    for required_option in REQUIRED_PREDICT_OPTIONS:
        if required_option not in predict_section:
            raise KeyError(
                f"the '{required_option}' option is required but was not found in the "
                f"PREDICT section of the config.toml file: {config_path}"
            )
    return PredictConfig(**predict_section)



"""constants used by multiple modules.
Defined here to avoid circular imports.
"""
from functools import partial

import crowsetta
import numpy as np
from evfuncs import load_cbin
from scipy.io import wavfile, loadmat

# ---- audio files ----
def swap_return_tuple_elements(func):
    def new_f(*args, **kwargs):
        return_tuple = func(*args, **kwargs)
        return return_tuple[1], return_tuple[0]
    return new_f


load_cbin = swap_return_tuple_elements(load_cbin)
AUDIO_FORMAT_FUNC_MAP = {
    'cbin': load_cbin,
    'wav': wavfile.read
}

VALID_AUDIO_FORMATS = list(AUDIO_FORMAT_FUNC_MAP.keys())

# ---- spectrogram files ----
SPECT_FORMAT_LOAD_FUNCTION_MAP = {
    'mat': partial(loadmat, squeeze_me=True),
    'npz': np.load,
}
VALID_SPECT_FORMATS = list(SPECT_FORMAT_LOAD_FUNCTION_MAP.keys())

# ---- annotation files ----
VALID_ANNOT_FORMATS = crowsetta.formats._INSTALLED
NO_ANNOTATION_FORMAT = 'none'



from pathlib import Path
import shutil
from datetime import datetime

from .. import config
from .. import core
from .. import logging


def train(toml_path):
    """train models using training set specified in config.toml file.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None

    Trains models, saves results in new directory within root_results_dir specified
    in config.toml file, and adds path to that new directory to config.toml file.
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml(toml_path)

    if cfg.train is None:
        raise ValueError(
            f'train called with a config.toml file that does not have a TRAIN section: {toml_path}'
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    results_dirname = f'results_{timenow}'
    if cfg.train.root_results_dir:
        results_path = Path(cfg.train.root_results_dir)
    else:
        results_path = Path('.')
    results_path = results_path.joinpath(results_dirname)
    results_path.mkdir(parents=True)
    # copy config file into results dir now that we've made the dir
    shutil.copy(toml_path, results_path)

    # ---- set up logging ----------------------------------------------------------------------------------------------
    logger = logging.get_logger(log_dst=results_path,
                                caller='train',
                                timestamp=timenow,
                                logger_name=__name__)
    logger.info('Logging results to {}'.format(results_path))

    model_config_map = config.models.map_from_path(toml_path, cfg.train.models)

    core.train(model_config_map,
               cfg.train.csv_path,
               cfg.prep.labelset,
               cfg.dataloader.window_size,
               cfg.train.batch_size,
               cfg.train.num_epochs,
               cfg.train.num_workers,
               results_path=results_path,
               spect_key=cfg.spect_params.spect_key,
               timebins_key=cfg.spect_params.timebins_key,
               normalize_spectrograms=cfg.train.normalize_spectrograms,
               shuffle=cfg.train.shuffle,
               val_step=cfg.train.val_step,
               ckpt_step=cfg.train.ckpt_step,
               patience=cfg.train.patience,
               device=cfg.train.device,
               logger=logger,
               )



from . import learncurve
from .annot import annotation, plot_labels, plot_segments
from .spect import spect, spect_annot

__all__ = [
    'learncurve',
    'plot_labels',
    'plot_segments',
    'spect',
    'spect_annot',
]



"""parses [PREP] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of

from .converters import expanded_user_path, labelset_from_toml_value
from .validators import is_a_directory, is_a_file, is_audio_format, is_annot_format, is_spect_format


def duration_from_toml_value(value):
    """converter for dataset split durations.
    If value is -1, that value is returned -- specifies "use the remainder of the dataset".
    Other values are converted to float when possible."""
    if value == -1:
        return value
    else:
        return float(value)


def is_valid_duration(instance, attribute, value):
    """validator for dataset split durations"""
    if type(value) not in {int, float}:
        raise TypeError(
            f'invalid type for {attribute} of {instance}: {type(value)}. Type should be float or int.'
        )

    if value == -1:  # specifies "use the remainder of the dataset"
        # so it is valid, but other negative values are not
        return

    if not value >= 0:
        raise ValueError(
            f'value specified for {attribute} of {instance} must be greater than or equal to zero, was {value}'
        )


@attr.s
class PrepConfig:
    """class to represent [PREP] section of config.toml file

    Attributes
    ----------
    data_dir : str
        path to directory with files from which to make dataset
    output_dir : str
        Path to location where data sets should be saved. Default is None,
        in which case data sets are saved in the current working directory.
    audio_format : str
        format of audio files. One of {'wav', 'cbin'}.
    spect_format : str
        format of files containg spectrograms as 2-d matrices.
        One of {'mat', 'npy'}.
    annot_format : str
        format of annotations. Any format that can be used with the
        crowsetta library is valid.
    annot_file : str
        Path to a single annotation file. Default is None.
        Used when a single file contains annotations for multiple audio files.
    labelset : set
        of str or int, the set of labels that correspond to annotated segments
        that a network should learn to segment and classify. Note that if there
        are segments that are not annotated, e.g. silent gaps between songbird
        syllables, then `vak` will assign a dummy label to those segments
        -- you don't have to give them a label here.
    train_dur : float
        total duration of training set, in seconds. When creating a learning curve,
        training subsets of shorter duration (specified by the 'train_set_durs' option
        in the LEARNCURVE section of a config.toml file) will be drawn from this set.
    val_dur : float
        total duration of validation set, in seconds.
    test_dur : float
        total duration of test set, in seconds.
    """
    data_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory)
    output_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory)

    audio_format = attr.ib(validator=validators.optional(is_audio_format), default=None)
    spect_format = attr.ib(validator=validators.optional(is_spect_format), default=None)
    annot_file = attr.ib(converter=converters.optional(expanded_user_path),
                         validator=validators.optional(is_a_file), default=None)
    annot_format = attr.ib(validator=validators.optional(is_annot_format), default=None)

    labelset = attr.ib(converter=converters.optional(labelset_from_toml_value),
                       validator=validators.optional(instance_of(set)),
                       default=None)

    train_dur = attr.ib(converter=converters.optional(duration_from_toml_value),
                        validator=validators.optional(is_valid_duration),
                        default=None)
    val_dur = attr.ib(converter=converters.optional(duration_from_toml_value),
                      validator=validators.optional(is_valid_duration),
                      default=None)
    test_dur = attr.ib(converter=converters.optional(duration_from_toml_value),
                       validator=validators.optional(is_valid_duration),
                       default=None)


REQUIRED_PREP_OPTIONS = [
    'data_dir',
]


def parse_prep_config(config_toml, config_path):
    """parse [PREP] section of config.toml file

    Parameters
    ----------
    config_toml : dict
        containing configuration file in TOML format, already loaded by parse function
    toml_path : Path
        path to a configuration file in TOML format (used for error messages)

    Returns
    -------
    prep_config : vak.config.prep.PrepConfig
        instance of class that represents [PREP] section of config.toml file
    """
    prep_section = config_toml['PREP']

    if 'spect_format' in prep_section and 'audio_format' in prep_section:
        raise ValueError("[PREP] section of config.toml file cannot specify both audio_format and "
                         "spect_format, unclear whether to create spectrograms from audio files or "
                         "use already-generated spectrograms")

    if 'spect_format' not in prep_section and 'audio_format' not in prep_section:
        raise ValueError("[PREP] section of config.toml file must specify either audio_format or "
                         "spect_format")

    for required_option in REQUIRED_PREP_OPTIONS:
        if required_option not in prep_section:
            raise KeyError(
                f"the '{required_option}' option is required but was not found in the "
                f"PREP section of the config.toml file: {config_path}"
            )

    return PrepConfig(**prep_section)



import json
import os
from pathlib import Path

import crowsetta
import joblib
import numpy as np
import pandas as pd
from tqdm import tqdm
import torch.utils.data

from .. import files
from .. import io
from .. import labeled_timebins
from ..logging import log_or_print
from .. import models
from .. import transforms
from ..datasets import VocalDataset
from ..device import get_default as get_default_device


def predict(csv_path,
            checkpoint_path,
            labelmap_path,
            model_config_map,
            window_size,
            num_workers=2,
            spect_key='s',
            timebins_key='t',
            spect_scaler_path=None,
            device=None,
            annot_csv_filename=None,
            output_dir=None,
            min_segment_dur=None,
            majority_vote=False,
            logger=None,
            ):
    """make predictions on dataset with trained model specified in config.toml file.
    Function called by command-line interface.

    Parameters
    ----------
    csv_path : str
        path to where dataset was saved as a csv.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    labelmap_path : str
        path to 'labelmap.json' file.
    model_config_map : dict
        where each key-value pair is model name : dict of config parameters
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shown to neural networks
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    device : str
        Device on which to work with model + data.
        Defaults to 'cuda' if torch.cuda.is_available is True.
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    annot_csv_filename : str
        name of .csv file containing predicted annotations.
        Default is None, in which case the name of the dataset .csv
        is used, with '.annot.csv' appended to it.
    output_dir : str, Path
        path to location where .csv containing predicted annotation
        should be saved. Defaults to current working directory.
    min_segment_dur : float
        minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from lbl_tb. Default is None, in which case no
        segments are removed.
    majority_vote : bool
        if True, transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment. This transform can only be
        applied if the labelmap contains an 'unlabeled' label,
        because unlabeled segments makes it possible to identify
        the labeled segments. Default is False.

    Other Parameters
    ----------------
    logger : logging.Logger
        instance created by vak.logging.get_logger. Default is None.

    Returns
    -------
    None
    """
    if output_dir is None:
        output_dir = Path(os.getcwd())
    else:
        output_dir = Path(output_dir)

    if not output_dir.is_dir():
        raise NotADirectoryError(
            f'value specified for output_dir is not recognized as a directory: {output_dir}'
        )

    if device is None:
        device = get_default_device()

    # ---------------- load data for prediction ------------------------------------------------------------------------
    if spect_scaler_path:
        log_or_print(f'loading SpectScaler from path: {spect_scaler_path}', logger=logger, level='info')
        spect_standardizer = joblib.load(spect_scaler_path)
    else:
        log_or_print(f'Not loading SpectScaler, no path was specified', logger=logger, level='info')
        spect_standardizer = None

    item_transform = transforms.get_defaults('predict',
                                             spect_standardizer,
                                             window_size=window_size,
                                             return_padding_mask=True,
                                             )

    log_or_print(f'loading labelmap from path: {labelmap_path}',
                 logger=logger, level='info')
    with labelmap_path.open('r') as f:
        labelmap = json.load(f)

    log_or_print(f'loading dataset to predict from csv path: {csv_path}', logger=logger, level='info')
    pred_dataset = VocalDataset.from_csv(csv_path=csv_path,
                                         split='predict',
                                         labelmap=labelmap,
                                         spect_key=spect_key,
                                         timebins_key=timebins_key,
                                         item_transform=item_transform,
                                         )

    pred_data = torch.utils.data.DataLoader(dataset=pred_dataset,
                                            shuffle=False,
                                            # batch size 1 because each spectrogram reshaped into a batch of windows
                                            batch_size=1,
                                            num_workers=num_workers)

    # ---------------- set up to convert predictions to annotation files -----------------------------------------------
    if annot_csv_filename is None:
        annot_csv_filename = Path(csv_path).stem + '.annot.csv'
    annot_csv_path = Path(output_dir).joinpath(annot_csv_filename)
    log_or_print(f'will save annotations in .csv file: {annot_csv_path}',
                 logger=logger, level='info')

    dataset_df = pd.read_csv(csv_path)
    timebin_dur = io.dataframe.validate_and_get_timebin_dur(dataset_df)
    log_or_print(f'dataset has timebins with duration: {timebin_dur}',
                 logger=logger, level='info')

    # ---------------- do the actual predicting + converting to annotations --------------------------------------------
    input_shape = pred_dataset.shape
    # if dataset returns spectrogram reshaped into windows,
    # throw out the window dimension; just want to tell network (channels, height, width) shape
    if len(input_shape) == 4:
        input_shape = input_shape[1:]
    log_or_print(f'shape of input to networks used for predictions: {input_shape}',
                 logger=logger, level='info')

    log_or_print(f'instantiating models from model-config map:/n{model_config_map}',
                 logger=logger, level='info')
    models_map = models.from_model_config_map(
        model_config_map,
        num_classes=len(labelmap),
        input_shape=input_shape
    )
    for model_name, model in models_map.items():
        # ---------------- do the actual predicting --------------------------------------------------------------------
        log_or_print(f'loading checkpoint for {model_name} from path: {checkpoint_path}',
                     logger=logger, level='info')
        model.load(checkpoint_path)
        log_or_print(f'running predict method of {model_name}',
                     logger=logger, level='info')
        pred_dict = model.predict(pred_data=pred_data,
                                  device=device)

        # ----------------  converting to annotations ------------------------------------------------------------------
        progress_bar = tqdm(pred_data)

        annots = []
        log_or_print('converting predictions to annotations',
                     logger=logger, level='info')
        for ind, batch in enumerate(progress_bar):
            padding_mask, spect_path = batch['padding_mask'], batch['spect_path']
            padding_mask = np.squeeze(padding_mask)
            if isinstance(spect_path, list) and len(spect_path) == 1:
                spect_path = spect_path[0]
            y_pred = pred_dict[spect_path]
            y_pred = torch.argmax(y_pred, dim=1)  # assumes class dimension is 1
            y_pred = torch.flatten(y_pred).cpu().numpy()[padding_mask]

            spect_dict = files.spect.load(spect_path)
            t = spect_dict[timebins_key]
            labels, onsets_s, offsets_s = labeled_timebins.lbl_tb2segments(y_pred,
                                                                           labelmap=labelmap,
                                                                           t=t,
                                                                           min_segment_dur=min_segment_dur,
                                                                           majority_vote=majority_vote)
            seq = crowsetta.Sequence.from_keyword(labels=labels,
                                                  onsets_s=onsets_s,
                                                  offsets_s=offsets_s)

            audio_fname = files.spect.find_audio_fname(spect_path)
            annot = crowsetta.Annotation(seq=seq, audio_file=audio_fname, annot_file=annot_csv_path.name)
            annots.append(annot)

        crowsetta.csv.annot2csv(annot=annots,
                                csv_filename=annot_csv_path)



import logging
import os

import numpy as np
import dask.bag as db
from dask.diagnostics import ProgressBar

from .. import constants
from .. import files
from ..annotation import source_annot_map
from ..config.spect_params import SpectParamsConfig
from ..spect import spectrogram


def files_from_dir(audio_dir, audio_format):
    """get all audio files of a given format
    from a directory or its sub-directories,
    using the file extension associated with that annotation format.

    Parameters
    ----------
    audio_dir : str
        path to directory containing audio files.
    audio_format : str
        valid audio file format. One of {'wav', 'cbin'}.

    Returns
    -------
    audio_files : list
        of paths to audio files
    """
    if audio_format not in constants.AUDIO_FORMAT_FUNC_MAP:
        raise ValueError(f"'{audio_format}' is not a valid audio format")
    audio_files = files.from_dir(audio_dir, audio_format)
    return audio_files


def to_spect(audio_format,
             spect_params,
             output_dir,
             audio_dir=None,
             audio_files=None,
             annot_list=None,
             audio_annot_map=None,
             labelset=None):
    """makes spectrograms from audio files and saves in array files

    Parameters
    ----------
    audio_format : str
        format of audio files. One of {'wav', 'cbin'}
    spect_params : dict or vak.config.spectrogram.SpectConfig
        parameters for computing spectrogram, from .toml file.
        To see all related parameters, run:
        >>> help(vak.config.spect_params.SpectParamConfig)
        To get a default configuration, create a SpectParamConfig
        with no arguments and then pass that to `to_spect`:
        >>> default_spect_params = vak.config.spect_params.SpectParamConfig()
        >>> to_spect(audio_format='wav', spect_params=default_spect_params, output_dir='.')
    audio_dir : str
        path to directory containing audio files from which to make spectrograms
    audio_files : list
        of str, full paths to audio files from which to make spectrograms
    annot_list : list
        of annotations for array files. Default is None.
    audio_annot_map : dict
        Where keys are paths to array files and value corresponding to each key is
        the annotation for that array file.
        Default is None.
    output_dir : str
        directory in which to save .spect.npz file generated for each audio file.
    labelset : set
        of str or int, set of unique labels for vocalizations. Default is None.
        If not None, then files will be skipped where the 'labels' array in the
        corresponding annotation contains labels that are not found in labelset

    Returns
    -------
    spect_files : list
        of str, full paths to .spect.npz files

    Notes
    -----
    For each audio file, a corresponding 'spect.npz' file will be created.
    Each '.spect.npz' file contains the following arrays:
        s : numpy.ndarray
            spectrogram, a 2-d array
        f : numpy.ndarray
            vector of centers of frequency bins from spectrogram
        t : numpy.ndarray
            vector of centers of tme bins from spectrogram
        audio_path : numpy.ndarray
            path to source audio file used to create spectrogram

    The names of the arrays are defaults, and will change if different values are specified
    in spect_params for 'spect_key', 'freqbins_key', 'timebins_key', or 'audio_path_key'.
    """
    if audio_format not in constants.VALID_AUDIO_FORMATS:
        raise ValueError(
            f"audio format must be one of '{constants.VALID_AUDIO_FORMATS}'; "
            f"format '{audio_format}' not recognized."
        )

    if all([arg is None for arg in (audio_dir, audio_files, audio_annot_map)]):
        raise ValueError('must specify one of: audio_dir, audio_files, audio_annot_map')

    if audio_dir and audio_files:
        raise ValueError('received values for audio_dir and audio_files, unclear which to use')

    if audio_dir and audio_annot_map:
        raise ValueError('received values for audio_dir and audio_annot_map, unclear which to use')

    if audio_files and audio_annot_map:
        raise ValueError('received values for audio_files and audio_annot_map, unclear which to use')

    if annot_list and audio_annot_map:
        raise ValueError(
            'received values for annot_list and array_annot_map, unclear which annotations to use'
        )

    if labelset is not None:
        if type(labelset) != set:
            raise TypeError(
                f'type of labelset must be set, but was: {type(labelset)}'
                f'type of labelset must be set, but was: {type(labelset)}'
            )

    if type(spect_params) not in [dict, SpectParamsConfig]:
        raise TypeError(
            'type of spect_params must be an instance of vak.config.spect_params.SpectParamsConfig, '
            'or a dict that can be converted to a SpectParamsConfig instance, '
            f'but was {type(spect_params)}'
        )
    if type(spect_params) is dict:
        spect_params = SpectParamsConfig(**spect_params)

    # validate audio files if supplied by user
    if audio_files:
        # make sure audio files are all the same type, and the same as audio format specified
        exts = []
        for audio_file in audio_files:
            root, ext = os.path.splitext(audio_file)
            exts.append(ext)
        uniq_ext = set(exts)
        if len(uniq_ext) > 1:
            raise ValueError(
                'audio_files should all have the same extension, '
                f'but found more than one: {uniq_ext}'
                )
        else:
            ext_str = uniq_ext.pop()
            if audio_format not in ext_str:
                raise ValueError(
                    f"audio format. '{audio_format}', does not match extensions in audio_files, '{ext_str}''"
                )

        if annot_list:  # make map we can validate below
            audio_annot_map = source_annot_map(audio_files, annot_list)

    # otherwise get audio files using audio dir (won't need to validate audio files)
    if audio_dir:
        audio_files = files_from_dir(audio_dir, audio_format)
        if annot_list:
            audio_annot_map = source_annot_map(audio_files, annot_list)

    logger = logging.getLogger(__name__)
    logger.setLevel('INFO')
    logger.info('creating array files with spectrograms')

    # use mapping (if generated/supplied) with labelset, if supplied, to filter
    if audio_annot_map:
        if labelset:  # then remove annotations with labels not in labelset
            # note we do this here so it happens regardless of whether
            # user supplied audio_annot_map *or* we constructed it above
            for audio_file, annot in list(audio_annot_map.items()):
                # loop in a verbose way (i.e. not a comprehension)
                # so we can give user warning when we skip files
                annot_labelset = set(annot.seq.labels)
                # below, set(labels_mapping) is a set of that dict's keys
                if not annot_labelset.issubset(set(labelset)):
                    # because there's some label in labels that's not in labelset
                    audio_annot_map.pop(audio_file)
                    logger.info(
                        f'found labels in {annot.annot_file} not in labels_mapping, '
                        f'skipping audio file: {audio_file}'
                    )
        audio_files = sorted(list(audio_annot_map.keys()))

    # this is defined here so all other arguments to 'to_spect' are in scope
    def _spect_file(audio_file):
        """helper function that enables parallelized creation of array
        files containing spectrograms.
        Accepts path to audio file, saves .npz file with spectrogram"""
        fs, dat = constants.AUDIO_FORMAT_FUNC_MAP[audio_format](audio_file)
        s, f, t = spectrogram(dat, fs,
                              spect_params.fft_size,
                              spect_params.step_size,
                              spect_params.thresh,
                              spect_params.transform_type,
                              spect_params.freq_cutoffs)
        spect_dict = {spect_params.spect_key: s,
                      spect_params.freqbins_key: f,
                      spect_params.timebins_key: t,
                      spect_params.audio_path_key: audio_file}
        basename = os.path.basename(audio_file)
        npz_fname = os.path.join(os.path.normpath(output_dir),
                                 basename + '.spect.npz')
        np.savez(npz_fname, **spect_dict)
        return npz_fname

    bag = db.from_sequence(audio_files)
    with ProgressBar():
        spect_files = list(bag.map(_spect_file))
    # sort because ordering from Dask not guaranteed
    spect_files = sorted(spect_files)
    return spect_files



import pandas as pd

from .. import annotation
from .. import files
from .. import labeled_timebins


class VocalDataset:
    """Base class to represent a dataset of vocalizations.
    It assumes that the dataset consists of spectrograms of
    vocalizations and, optionally, annotations
    for those vocalizations.
    """
    def __init__(self,
                 csv_path,
                 spect_paths,
                 annots=None,
                 labelmap=None,
                 spect_key='s',
                 timebins_key='t',
                 item_transform=None,
                 ):
        """initialize a VocalDataset instance

        Parameters
        ----------
        csv_path : str, Path
            path to a .csv file that represents the dataset.
        spect_paths : numpy.ndarray
            column from DataFrame that represents dataset,
            consisting of paths to files containing spectrograms as arrays
        annots : list
            of crowsetta.Annotation instances,
            loaded from from DataFrame that represents dataset, using vak.annotation.from_df.
            Default is None, in which case no annotation is returned with each item
            in the dataset.
        labelmap : dict
            that maps labels from dataset to a series of consecutive integer.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        item_transform : callable
            A function / transform that takes an input numpy array or torch Tensor,
            and optionally a target array or Tensor, and returns a dictionary.
            This dictionary is the item returned when indexing into the dataset.
            Default is None.
        """
        self.csv_path = csv_path
        self.spect_paths = spect_paths
        self.spect_key = spect_key
        self.timebins_key = timebins_key
        self.annots = annots
        self.labelmap = labelmap
        if 'unlabeled' in self.labelmap:
            self.unlabeled_label = self.labelmap['unlabeled']
        else:
            # if there is no "unlabeled label" (e.g., because all segments have labels)
            # just assign dummy value that will end up getting replaced by actual labels by label_timebins()
            self.unlabeled_label = 0
        self.item_transform = item_transform

        tmp_x_ind = 0
        tmp_item = self.__getitem__(tmp_x_ind)
        # used by vak functions that need to determine size of input,
        # e.g. when initializing a neural network model
        self.shape = tmp_item['source'].shape

    def __getitem__(self, idx):
        spect_path = self.spect_paths[idx]
        spect_dict = files.spect.load(spect_path)
        spect = spect_dict[self.spect_key]

        if self.annots is not None:
            timebins = spect_dict[self.timebins_key]

            annot = self.annots[idx]
            lbls_int = [self.labelmap[lbl] for lbl in annot.seq.labels]
            # "lbl_tb": labeled timebins. Target for output of network
            lbl_tb = labeled_timebins.label_timebins(lbls_int,
                                                     annot.seq.onsets_s,
                                                     annot.seq.offsets_s,
                                                     timebins,
                                                     unlabeled_label=self.unlabeled_label)
            item = self.item_transform(spect, lbl_tb, spect_path)
        else:
            item = self.item_transform(spect, spect_path)

        return item

    def __len__(self):
        """number of batches"""
        return len(self.spect_paths)

    @classmethod
    def from_csv(cls, csv_path, split, labelmap,
                 spect_key='s', timebins_key='t', item_transform=None):
        """given a path to a csv representing a dataset,
        returns an initialized VocalDataset.

        Parameters
        ----------
        csv_path : str, Path
            path to a .csv file that represents the dataset.
        spect_paths : numpy.ndarray
            column from DataFrame that represents dataset,
            consisting of paths to files containing spectrograms as arrays
        annots : list
            of crowsetta.Annotation instances,
            loaded from from DataFrame that represents dataset, using vak.annotation.from_df.
            Default is None, in which case no annotation is returned with each item
            in the dataset.
        labelmap : dict
            that maps labels from dataset to a series of consecutive integer.
            To create a label map, pass a set of labels to the `vak.utils.labels.to_map` function.
        spect_key : str
            key to access spectograms in array files. Default is 's'.
        timebins_key : str
            key to access time bin vector in array files. Default is 't'.
        item_transform : callable
            A function / transform that takes an input numpy array or torch Tensor,
            and optionally a target array or Tensor, and returns a dictionary.
            This dictionary is the item returned when indexing into the dataset.
            Default is None.

        Returns
        -------
        initialized instance of VocalDataset
        """
        df = pd.read_csv(csv_path)
        if not df['split'].str.contains(split).any():
            raise ValueError(
                f'split {split} not found in dataset in csv: {csv_path}'
            )
        else:
            df = df[df['split'] == split]

        spect_paths = df['spect_path'].values

        # below, annots will be None if no format is specified in the `annot_format` column of the dataframe.
        # this is intended behavior; makes it possible to use same dataset class for prediction
        annots = annotation.from_df(df)

        return cls(csv_path,
                   spect_paths,
                   annots,
                   labelmap,
                   spect_key,
                   timebins_key,
                   item_transform,
                   )



from .__about__ import (
    __author__,
    __commit__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
)

from . import __main__
from . import annotation
from . import cli
from . import config
from . import csv
from . import datasets
from . import device
from . import engine
from . import entry_points
from . import files
from . import io
from . import labels
from . import labeled_timebins
from . import logging
from . import metrics
from . import models
from . import plot
from . import spect
from . import summary_writer
from . import timebins
from . import transforms
from . import split
from . import validation

from .engine.model import Model


__all__ = [
    '__main__',
    'annotation',
    'cli',
    'config',
    'csv',
    'datasets',
    'device',
    'engine',
    'entry_points',
    'files',
    'io',
    'labels',
    'labeled_timebins',
    'logging',
    'metrics',
    'Model',
    'models',
    'plot',
    'spect',
    'split',
    'summary_writer',
    'timebins',
    'transforms',
    'validation',
]



import random
import itertools
import logging

from .validate import validate_durations_convert_nonnegative


def brute_force(durs,
                labels,
                labelset,
                train_dur,
                val_dur,
                test_dur,
                max_iter=5000):
    """finds indices that split (labels, durations) tuples into training,
    test, and validation sets of specified durations, with the set of unique labels
    in each dataset equal to the specified labelset.

    The durations of the datasets created using the returned indices will be
    *greater than* or equal to the durations specified.

    Must specify a positive value for one of {train_dur, test_dur}.
    The other value can be specified as '-1' which is interpreted as "use the
    remainder of the dataset for this split after finding indices for the set with a specified duration".
    If only one of {train_dur, test_dur} is specified, the other defaults to '-1'.

    Parameters
    ----------
    durs : list
        of durations of vocalizations
    labels : list
        of labels from vocalizations
    labelset : set
        of labels
    train_dur : int, float
        Target duration for training set, in seconds.
    val_dur : int, float
        Target duration for validation set, in seconds.
    test_dur : int, float
        Target duration for test set, in seconds.
    max_iter : int
        maximum number of iterations to attempt to find indices. Default is 5000.

    Returns
    -------
    train_inds, val_inds, test_inds : list
        of int, the indices that will split datasets

    Notes
    -----
    A 'brute force' algorithm that just randomly assigns indices to a set,
    and iterates until it finds some partition where each set has instances of all classes of label.
    Starts by ensuring that each label is represented in each set and then adds files to reach the required
    durations.
    """
    logger = logging.getLogger(__name__)
    logger.setLevel('INFO')

    sum_durs = sum(durs)
    train_dur, val_dur, test_dur = validate_durations_convert_nonnegative(train_dur, val_dur, test_dur, sum_durs)
    total_target_dur = sum([dur for dur in (train_dur, val_dur, test_dur) if dur is not None])

    durs_labels_list = list(zip(durs, labels))
    iter = 1
    all_labels_err = ('Did not successfully divide data into training, '
                      'validation, and test sets of sufficient duration '
                      f'after {max_iter} iterations.'
                      ' Try increasing the total size of the data set.')

    # ---- outer loop that repeats until we successfully split our reach max number of iters ---------------------------
    while 1:
        train_inds = []
        test_inds = []
        if val_dur > 0:
            val_inds = []
        else:
            val_inds = None

        total_train_dur = 0
        total_val_dur = 0
        total_test_dur = 0

        durs_labels_inds = list(range(len(durs_labels_list)))

        choice = []
        if train_dur > 0:
            choice.append('train')
            lset_train = set()
        else:
            lset_train = None
        if test_dur > 0:
            choice.append('test')
            lset_test = set()
        else:
            lset_test = None
        if val_dur:
            choice.append('val')
            lset_val = set()
        else:
            lset_val = None

        # ---- make sure each split has at least one instance of each label
        for label in sorted(labelset):
            label_inds = [ind for ind, labels in enumerate(labels)
                          if label in labels
                          and ind in durs_labels_inds]
            if len(label_inds) < len(choice):
                raise ValueError(
                    f'unable to split dataset so that each split has an instance of label: {label}.'
                    f'There were only {len(label_inds)} files with that label, '
                    f'but there are {len(choice)} splits.'
                )

            random.shuffle(label_inds)
            if train_dur > 0 and label not in lset_train:
                try:
                    ind = label_inds.pop()
                    train_inds.append(ind)
                    total_train_dur += durs[ind]
                    lset_train = lset_train.union(set(labels[ind]))
                    durs_labels_inds.remove(ind)
                except IndexError:
                    if len(label_inds) == 0:
                        logger.debug(
                            'Ran out of elements while dividing dataset into subsets of specified durations.'
                            f'Iteration {iter}'
                        )
                        iter += 1
                        break  # do next iteration
                    else:
                        # something else happened, re-raise error
                        raise

            if test_dur > 0 and label not in lset_test:
                try:
                    ind = label_inds.pop()
                    test_inds.append(ind)
                    total_test_dur += durs[ind]
                    lset_test = lset_test.union(set(labels[ind]))
                    durs_labels_inds.remove(ind)
                except IndexError:
                    if len(label_inds) == 0:
                        logger.debug(
                            'Ran out of elements while dividing dataset into subsets of specified durations.'
                            f'Iteration {iter}'
                        )
                        iter += 1
                        break  # do next iteration
                    else:
                        # something else happened, re-raise error
                        raise

            if val_dur > 0 and label not in lset_val:
                try:
                    ind = label_inds.pop()
                    val_inds.append(ind)
                    total_val_dur += durs[ind]
                    lset_val = lset_val.union(set(labels[ind]))
                    durs_labels_inds.remove(ind)
                except IndexError:
                    if len(label_inds) == 0:
                        logger.debug(
                            'Ran out of elements while dividing dataset into subsets of specified durations.'
                            f'Iteration {iter}'
                        )
                        iter += 1
                        break  # do next iteration
                    else:
                        # something else happened, re-raise error
                        raise

        random.shuffle(durs_labels_inds)
        if train_dur > 0 and total_train_dur >= train_dur:
            choice.remove('train')
        if test_dur > 0 and total_test_dur >= test_dur:
            choice.remove('test')
        if val_dur > 0 and total_val_dur >= val_dur:
            choice.remove('val')

        if len(choice) == 0:
            finished = True
        else:
            finished = False

        # ---- inner loop that actually does split -----------------
        while finished is False:
            # pop durations off list and append to randomly-chosen
            # list, either train, val, or test set.
            # Do this until the total duration for each data set is equal
            # to or greater than the target duration for each set.
            try:
                ind = durs_labels_inds.pop()
            except IndexError:
                if len(durs_labels_inds) == 0:
                    logger.debug(
                        'Ran out of elements while dividing dataset into subsets of specified durations.'
                        f'Iteration {iter}'
                    )
                    iter += 1
                    break  # do next iteration
                else:
                    # something else happened, re-raise error
                    raise

            which_set = random.randint(0, len(choice) - 1)
            which_set = choice[which_set]
            if which_set == 'train':
                train_inds.append(ind)
                total_train_dur += durs_labels_list[ind][0]  # ind 0 is duration
                if total_train_dur >= train_dur:
                    choice.pop(choice.index('train'))
            elif which_set == 'val':
                val_inds.append(ind)
                total_val_dur += durs_labels_list[ind][0]  # ind 0 is duration
                if total_val_dur >= val_dur:
                    choice.pop(choice.index('val'))
            elif which_set == 'test':
                test_inds.append(ind)
                total_test_dur += durs_labels_list[ind][0]  # ind 0 is duration
                if total_test_dur >= test_dur:
                    choice.pop(choice.index('test'))

            if len(choice) < 1:
                total_all_durs = sum([dur for dur in (total_train_dur, total_val_dur, total_test_dur)
                                      if dur is not None])
                if total_all_durs < total_target_dur:
                    raise ValueError(
                        f'Loop to find subsets completed but total duration of subsets, {total_all_durs} seconds, '
                        f'is less than total duration specified: {total_target_dur} seconds.')
                else:
                    finished = True
                    break

        if iter > max_iter:
            raise ValueError('Could not find subsets of sufficient duration in '
                             f'less than {max_iter} iterations.')

        if finished is True:
            # make sure that each set contains all classes we
            # want the network to learn
            if train_dur > 0:
                train_tups = [durs_labels_list[ind] for ind in train_inds]  # tup = a tuple
                train_labels = itertools.chain.from_iterable(
                    [tup[1] for tup in train_tups])
                train_labelset = set(train_labels)  # make set to get unique values
                if train_labelset != set(labelset):
                    iter += 1
                    if iter > max_iter:
                        raise ValueError(all_labels_err)
                    else:
                        logger.debug(
                            'Train labels did not contain all labels in labelset. '
                            f'Getting new training set. Iteration {iter}'
                        )
                        continue

            if test_dur > 0:
                test_tups = [durs_labels_list[ind] for ind in test_inds]
                test_labels = itertools.chain.from_iterable(
                    [tup[1] for tup in test_tups])
                test_labelset = set(test_labels)
                if test_labelset != set(labelset):
                    iter += 1
                    if iter > max_iter:
                        raise ValueError(all_labels_err)
                    else:
                        logger.debug(
                            'Test labels did not contain all labels in labelset. '
                            f'Getting new test set. Iteration {iter}'
                        )
                        continue

            if val_dur > 0:
                val_tups = [durs_labels_list[ind] for ind in val_inds]
                val_labels = itertools.chain.from_iterable(
                    [tup[1] for tup in val_tups])
                val_labelset = set(val_labels)
                if val_labelset != set(labelset):
                    iter += 1
                    if iter > max_iter:
                        raise ValueError(all_labels_err)
                    else:
                        logger.debug(
                            'Validation labels did not contain all labels in labelset. '
                            f'Getting new validation set. Iteration {iter}'
                        )
                        continue

            # successfully split
            break

        elif finished is False:
            continue

    return train_inds, val_inds, test_inds



"""parses [EVAL] section of config"""
import attr
from attr import converters, validators
from attr.validators import instance_of

from .converters import comma_separated_list, expanded_user_path
from .validators import is_a_directory, is_a_file, is_valid_model_name
from .. import device


@attr.s
class EvalConfig:
    """class that represents [EVAL] section of config.toml file

    Attributes
    ----------
    csv_path : str
        path to where dataset was saved as a csv.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    output_dir : str
        Path to location where .csv files with evaluation metrics should be saved.
    labelmap_path : str
        path to 'labelmap.json' file.
    models : list
        of model names. e.g., 'models = TweetyNet, GRUNet, ConvNet'
    batch_size : int
        number of samples per batch presented to models during training.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    device : str
        Device on which to work with model + data.
        Defaults to 'cuda' if torch.cuda.is_available is True.
    spect_scaler_path : str
        path to a saved SpectScaler object used to normalize spectrograms.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    """
    # required, external files
    checkpoint_path = attr.ib(converter=expanded_user_path,
                              validator=is_a_file)
    labelmap_path = attr.ib(converter=expanded_user_path,
                            validator=is_a_file)
    output_dir = attr.ib(converter=expanded_user_path, validator=is_a_directory)

    # required, model / dataloader
    models = attr.ib(converter=comma_separated_list,
                     validator=[instance_of(list), is_valid_model_name])
    batch_size = attr.ib(converter=int, validator=instance_of(int))

    # csv_path is actually 'required' but we can't enforce that here because cli.prep looks at
    # what sections are defined to figure out where to add csv_path after it creates the csv
    csv_path = attr.ib(converter=converters.optional(expanded_user_path),
                       validator=validators.optional(is_a_file),
                       default=None
                       )

    # optional, transform
    spect_scaler_path = attr.ib(converter=converters.optional(expanded_user_path),
                                validator=validators.optional(is_a_file),
                                default=None)

    # optional, data loader
    num_workers = attr.ib(validator=instance_of(int), default=2)
    device = attr.ib(validator=instance_of(str), default=device.get_default())


REQUIRED_EVAL_OPTIONS = [
    'checkpoint_path',
    'labelmap_path',
    'output_dir',
    'models',
]


def parse_eval_config(config_obj, config_path):
    """parse [EVAL] section of config.toml file

    Parameters
    ----------
    config_obj : ConfigParser
        containing config.toml file already loaded by parse function
    config_path : str
        path to config.toml file (used for error messages)

    Returns
    -------
    predict_config : vak.config.predict.PredictConfig
        instance of PredictConfig class that represents [PREDICT] section
        of config.toml file
    """
    eval_section = dict(
        config_obj['EVAL'].items()
    )

    for required_option in REQUIRED_EVAL_OPTIONS:
        if required_option not in eval_section:
            raise KeyError(
                f"the '{required_option}' option is required but was not found in the "
                f"EVAL section of the config.toml file: {config_path}"
            )
    return EvalConfig(**eval_section)



"""
Invokes __main__ when the module is run as a script.
Example: python -m vak --help
"""
import argparse
from pathlib import Path

from .cli import cli


def get_parser():
    """returns ArgumentParser instance used by main()"""
    CHOICES = [
        'prep',
        'train',
        'eval',
        'predict',
        'finetune',
        'learncurve',
    ]

    parser = argparse.ArgumentParser(description='vak command-line interface',
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('command', type=str, metavar='command',
                        choices=CHOICES,
                        help="Command to run, valid options are:\n"
                             f"{CHOICES}\n"
                             "$ vak train ./configs/config_2018-12-17.toml")
    parser.add_argument('configfile', type=Path,
                        help='name of config.toml file to use \n'
                             '$ vak train ./configs/config_2018-12-17.toml')
    return parser


def main():
    """main function, called when package is run with `python -m vak` or
    alternatively just calling `vak` at the command line (because this
    function is installed under just `vak` as a console script entry point,
    see setup.py)
    """
    parser = get_parser()
    args = parser.parse_args()
    cli(command=args.command,
        config_file=args.configfile)


if __name__ == "__main__":
    main()



"""command-line interface functions for training,
creating learning curves, etc."""
from .prep import prep
from .learncurve import learning_curve
from .predict import predict
from .cli import cli
from .train import train



from datetime import datetime
from pathlib import Path

from .. import config
from .. import core
from .. import logging


def predict(toml_path):
    """make predictions on dataset with trained model specified in config.toml file.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml(toml_path)

    if cfg.predict is None:
        raise ValueError(
            f'predict called with a config.toml file that does not have a PREDICT section: {toml_path}'
        )

    # ---- set up logging ----------------------------------------------------------------------------------------------
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    logger = logging.get_logger(log_dst=cfg.prep.output_dir,
                                caller='predict',
                                timestamp=timenow,
                                logger_name=__name__)
    logger.info('Logging results to {}'.format(cfg.prep.output_dir))

    model_config_map = config.models.map_from_path(toml_path, cfg.predict.models)

    core.predict(csv_path=cfg.predict.csv_path,
                 checkpoint_path=cfg.predict.checkpoint_path,
                 labelmap_path=cfg.predict.labelmap_path,
                 model_config_map=model_config_map,
                 window_size=cfg.dataloader.window_size,
                 num_workers=cfg.predict.num_workers,
                 spect_key=cfg.spect_params.spect_key,
                 timebins_key=cfg.spect_params.timebins_key,
                 spect_scaler_path=cfg.predict.spect_scaler_path,
                 device=cfg.predict.device,
                 annot_csv_filename=cfg.predict.annot_csv_filename,
                 output_dir=cfg.predict.output_dir,
                 min_segment_dur=cfg.predict.min_segment_dur,
                 majority_vote=cfg.predict.majority_vote,
                 logger=logger)



from .eval import eval
from .train import train
from .learncurve import learning_curve
from .predict import predict
from .prep import prep


def cli(command, config_file):
    """command-line interface

    Parameters
    ----------
    command : string
        One of {'prep', 'train', 'eval', 'predict', 'finetune', 'learncurve'}
    config_file : str, Path
        path to a config.toml file
    """
    if command == 'prep':
        prep(toml_path=config_file)

    elif command == 'train':
        train(toml_path=config_file)

    elif command == 'eval':
        eval(toml_path=config_file)

    elif command == 'predict':
        predict(toml_path=config_file)

    elif command == 'learncurve':
        learning_curve(toml_path=config_file)

    elif command == 'finetune':
        raise NotImplementedError

    else:
        raise ValueError(
            f'command not recognized: {command}'
        )



"""validators used by attrs-based classes and by vak.parse.parse_config"""
from pathlib import Path

import toml

from .. import constants
from .. import models


def is_a_directory(instance, attribute, value):
    """check if given path is a directory"""
    if not Path(value).is_dir():
        raise NotADirectoryError(
            f'Value specified for {attribute.name} of {type(instance)} not recognized as a directory:\n'
            f'{value}'
        )


def is_a_file(instance, attribute, value):
    """check if given path is a file"""
    if not Path(value).is_file():
        raise FileNotFoundError(
            f'Value specified for {attribute.name} of {type(instance)} not recognized as a file:\n'
            f'{value}'
        )


def is_valid_model_name(instance, attribute, value):
    MODEL_NAMES = [model_name for model_name, model_builder in models.find()]
    for model_name in value:
        if model_name not in MODEL_NAMES and f'{model_name}Model' not in MODEL_NAMES:
            raise ValueError(
                f'Model {model_name} not found when importing installed models.'
            )


def is_audio_format(instance, attribute, value):
    """check if valid audio format"""
    if value not in constants.VALID_AUDIO_FORMATS:
        raise ValueError(
            f'{value} is not a valid format for audio files'
        )


def is_annot_format(instance, attribute, value):
    """check if valid annotation format"""
    if value not in constants.VALID_ANNOT_FORMATS:
        raise ValueError(
            f'{value} is not a valid format for annotation files.\n'
            f'Valid formats are: {constants.VALID_ANNOT_FORMATS}'
        )


def is_spect_format(instance, attribute, value):
    """check if valid format for spectrograms"""
    if value not in constants.VALID_SPECT_FORMATS:
        raise ValueError(
            f'{value} is not a valid format for spectrogram files.\n'
            f'Valid formats are: {constants.VALID_SPECT_FORMATS}'
        )


CONFIG_DIR = Path(__file__).parent
VALID_TOML_PATH = CONFIG_DIR.joinpath('valid.toml')
with VALID_TOML_PATH.open('r') as fp:
    VALID_DICT = toml.load(fp)
VALID_SECTIONS = list(VALID_DICT.keys())
VALID_OPTIONS = {
    section: list(options.keys())
    for section, options in VALID_DICT.items()
}


def are_sections_valid(config_dict, toml_path):
    sections = list(config_dict.keys())
    MODEL_NAMES = [model_name for model_name, model_builder in models.find()]
    # add model names to valid sections so users can define model config in sections
    valid_sections = VALID_SECTIONS + MODEL_NAMES
    for section in sections:
        if section not in valid_sections and f'{section}Model' not in valid_sections:
            raise ValueError(
                f'section defined in {toml_path} is not valid: {section}'
            )


def are_options_valid(config_dict, section, toml_path):
    user_options = set(config_dict[section].keys())
    valid_options = set(VALID_OPTIONS[section])
    if not user_options.issubset(valid_options):
        invalid_options = user_options - valid_options
        raise ValueError(
            f"the following options from {section} section in "
            f"the config file '{toml_path.name}' are not valid:\n{invalid_options}"
        )



from pathlib import Path
from datetime import datetime
import warnings

import toml

from .. import config
from .. import core
from .. import logging


def prep(toml_path):
    """prepare datasets from vocalizations.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.
        Used to rewrite file with options determined by this function and needed for other functions

    Returns
    -------
    None

    Notes
    -----
    Saves a .csv file representing the dataset generated from data_dir.

    Datasets are used to train neural networks that segment audio files into
    vocalizations, and then predict labels for those segments.
    The function also prepares datasets so neural networks can predict the
    segmentation and annotation of vocalizations in them.
    It can also split a dataset into training, validation, and test sets,
    e.g. for benchmarking different neural network architectures.

    If the 'purpose' is set to 'train' or 'learncurve', and/or
    the duration of either the training or test set is provided,
    then the function attempts to split the dataset into training and test sets.
    A duration can also be specified for a validation set
    (used to measure performance during training).
    In these cases, the 'split' column in the .csv
    identifies which files (rows) belong to the training, test, and
    validation sets created from that Dataset.

    If the 'purpose' is set to 'predict' or 'eval',
    or no durations for any of the training sets are specified,
    then the function assumes all the vocalizations constitute a single
    dataset, and for all rows the 'split' columns for that dataset
    will be 'predict' or 'test' (respectively).
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml(toml_path)

    if cfg.prep is None:
        raise ValueError(
            f'prep called with a config.toml file that does not have a PREP section: {toml_path}'
        )

    # ---- set up logging ----------------------------------------------------------------------------------------------
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    logger = logging.get_logger(log_dst=cfg.prep.output_dir,
                                caller='prep',
                                timestamp=timenow,
                                logger_name=__name__)

    # ---- figure out purpose of config file from sections; will save csv path in that section -------------------------
    with toml_path.open('r') as fp:
        config_toml = toml.load(fp)

    if 'EVAL' in config_toml:
        section = 'EVAL'
    elif 'LEARNCURVE' in config_toml:
        section = 'LEARNCURVE'
    elif 'PREDICT' in config_toml:
        section = 'PREDICT'
        if cfg.prep.labelset is not None:
            warnings.warn(
                "config has a PREDICT section, but labelset option is specified in PREP section."
                "This would cause an error because the dataframe.from_files section will attempt to "
                f"check whether the files in the data_dir ({cfg.prep.data_dir}) have labels in "
                "labelset, even though those files don't have annotation.\n"
                "Setting labelset to None."
            )
            cfg.prep.labelset = None
    elif 'TRAIN' in config_toml:
        section = 'TRAIN'
    else:
        raise ValueError(
            'Did not find a section named TRAIN, LEARNCURVE, or PREDICT in config.toml file;'
            ' unable to determine which section to add paths to prepared datasets to'
        )
    logger.info(f'determined that config file has section: {section}\nWill add csv_path option to that section')

    purpose = section.lower()
    vak_df, csv_path = core.prep(data_dir=cfg.prep.data_dir,
                                 purpose=purpose,
                                 audio_format=cfg.prep.audio_format,
                                 spect_format=cfg.prep.spect_format,
                                 spect_params=cfg.spect_params,
                                 annot_format=cfg.prep.annot_format,
                                 annot_file=cfg.prep.annot_file,
                                 labelset=cfg.prep.labelset,
                                 output_dir=cfg.prep.output_dir,
                                 train_dur=cfg.prep.train_dur,
                                 val_dur=cfg.prep.val_dur,
                                 test_dur=cfg.prep.test_dur,
                                 logger=logger,
                                 )

    # use config and section from above to add csv_path to config.toml file
    config_toml[section]['csv_path'] = str(csv_path)

    with toml_path.open('w') as fp:
        toml.dump(config_toml, fp)



import attr
from attr.validators import instance_of


@attr.s
class DataLoaderConfig:
    """represents options for DataLoaders specified in config.toml file

    Attributes
    ----------
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shonw to neural networks
    """
    window_size = attr.ib(converter=int,
                          validator=instance_of(int),
                          default=88)


def parse_dataloader_config(config, config_path):
    # return defaults if config doesn't have SPECTROGRAM section
    dataloader_section = {}
    if 'DATALOADER' in config:
        dataloader_section.update(
            config['DATALOADER'].items()
        )
    return DataLoaderConfig(**dataloader_section)



import os
import re
from glob import glob


def find_fname(fname, ext):
    """given a file extension, finds a filename with that extension within
    another filename. Useful to find e.g. names of audio files in
    names of spectrogram files.

    Parameters
    ----------
    fname : str
        filename to search for another filename with a specific extension
    ext : str
        extension to search for in filename

    Returns
    -------
    sub_fname : str or None


    Examples
    --------
    >>> sub_fname(fname='llb3_0003_2018_04_23_14_18_54.wav.mat', ext='wav')
    'llb3_0003_2018_04_23_14_18_54.wav'
    """
    m = re.match(f'[\S]*{ext}', fname)
    if hasattr(m, 'group'):
        return m.group()
    elif m is None:
        return m


def from_dir(dir_path, ext):
    """helper function that gets all files with a given extension
    from a directory or its sub-directories.

    If no files with the specified extension are found in the directory, then
    the function recurses into all sub-directories and returns any files with
    the extension in those sub-directories.

    Parameters
    ----------
    dir_path : str
        path to target directory
    ext : str
        file extension to search for

    Returns
    -------
    files : list
        of paths to files with specified file extension

    Notes
    -----
    used by vak.io.audio.files_from_dir and vak.io.annot.files_from_dir
    """
    wildcard_with_extension = f'*.{ext}'
    files = sorted(
        glob(os.path.join(dir_path, wildcard_with_extension))
    )
    if len(files) == 0:
        # if we don't any files with extension, look in sub-directories
        files = []
        subdirs = glob(os.path.join(dir_path, '*/'))
        for subdir in subdirs:
            files.extend(
                glob(os.path.join(dir_path, subdir, wildcard_with_extension))
            )

    if len(files) == 0:
        raise FileNotFoundError(
            f'No files with extension {ext} found in '
            f'{dir_path} or immediate sub-directories'
        )

    return files



"""utility functions dealing with tensorboard SummaryWriter"""
from torch.utils.tensorboard import SummaryWriter


def get_summary_writer(log_dir, filename_suffix):
    """get an instance of a SummaryWriter,
    to use with a vak.Model during training

    Parameters
    ----------
    log_dir : str
        directory where event file will be written
    filename_suffix : str
        suffix added to events file name

    Returns
    -------
    summary_writer : torch.utils.tensorboard.SummaryWriter

    Examples
    --------
    >>> summary_writer = vak.summary_writer.get_summary_writer(log_dir='./experiments')
    >>> tweety_net_model.summary_writer = summary_writer  # set attribute equal to instance we just made
    >>> tweety_net_model.train()  # now events during training will be logged with that summary writer
    """
    return SummaryWriter(
        log_dir=log_dir,
        filename_suffix=filename_suffix
    )



def validate_durations_convert_nonnegative(train_dur, val_dur, test_dur, vds_dur):
    """helper function to validate durations specified for splits,
    and convert all durations to non-negative numbers,
    so other functions can do the actual splitting.

    First the functions checks for invalid conditions:
        + If train_dur, val_dur, and test_dur are all None, a ValueError is raised.
        + If any of train_dur, val_dur, or test_dur have a negative value that is not -1, an
          ValueError is raised. -1 is interpreted differently as explained below.
        + If only val_dur is specified, this raises a ValueError; not clear what durations of training
          and test set should be.

    Then, if either train_dur or test_dur are None, they are set to 0. None means user did not specify a value.

    Finally the function validates that the sum of the specified split durations is not greater than
    the the total duration of the dataset, `vds_dur`.

    If any split is specified as -1, this value is interpreted as "first get the
    split for the set with a value specified, then use the remainder of the dataset in the split
    whose duration is set to -1". The duration for the split specified as -1 will be computed by
    first summing the value of the other two splits, validating that they are not larger than
    the total duration of the dataset, and finally subtracting off the combined duration of the
    other two.

    Parameters
    ----------
    train_dur : int, float
        Target duration for training set, in seconds.
    val_dur : int, float
        Target duration for validation set, in seconds.
    test_dur : int, float
        Target duration for test set, in seconds.
    vds_dur : int, float
        Total duration of Dataset.

    Returns
    -------
    train_dur, val_dur, test_dur : int, float
    """
    if val_dur and (train_dur is None and test_dur is None):
        raise ValueError(
            'cannot specify only val_dur, unclear how to split dataset into training and test sets'
        )

    # make a dict so we can refer to variable by name in loop
    split_durs = {
        'train': train_dur,
        'val': val_dur,
        'test': test_dur,
    }
    if all([dur is None for dur in split_durs.values()]):
        raise ValueError("train_dur, val_dur, and test_dur were all None; must specify at least train_dur or test_dur")

    if not all([dur >= 0 or dur == -1 for dur in split_durs.values() if dur is not None]):
        raise ValueError(
            "all durations for split must be real non-negative number or "
            "set to -1 (meaning 'use the remaining dataset)"
        )

    if sum([split_dur == -1 for split_dur in split_durs.values()]) > 1:
        raise ValueError(
            'cannot specify duration of more than one split as -1, unclear how to calculate durations of splits.'
        )

    # set any None values for durations to 0; no part of dataset will go to that split
    for split_name in split_durs.keys():
        if split_durs[split_name] is None:
            split_durs[split_name] = 0

    if -1 in split_durs.values():
        total_other_splits_dur = sum([dur for dur in split_durs.values() if dur is not -1])

        if total_other_splits_dur > vds_dur:
            raise ValueError(
                'One dataset split duration was specified as -1, but the total of the other durations specified, '
                f'{total_other_splits_dur} s, is greater than total duration of Dataset, {vds_dur}.'
            )
        else:
            remainder = vds_dur - total_other_splits_dur
            for split_name in split_durs.keys():
                if split_durs[split_name] == -1:
                    split_durs[split_name] = remainder

    # validate one last time now that we have real non-negative values for all split durations
    total_splits_dur = sum(split_durs.values())

    if total_splits_dur > vds_dur:
        raise ValueError(
            f'Total of the split durations specified, {total_splits_dur} s, '
            f'is greater than total duration of dataset, {vds_dur}.'
        )

    return split_durs['train'], split_durs['val'], split_durs['test']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# adopted from Kenneth Reitz
# https://github.com/kennethreitz/setup.py
# released under MIT license
# (https://github.com/kennethreitz/setup.py/blob/master/LICENSE)

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# Package meta-data.
about = {}
with open("src/vak/__about__.py") as fp:
    exec(fp.read(), about)

NAME = about['__title__']
DESCRIPTION = about['__summary__']
URL = about['__uri__']
EMAIL = about['__email__']
AUTHOR = about['__author__']
REQUIRES_PYTHON = '>=3.6.0'
VERSION = about['__version__']
LICENSE = about['__license__']

REQUIRED = [
    'attrs',
    'crowsetta>=2.2.0',
    'dask[bag]',
    'evfuncs',
    'joblib',
    'matplotlib',
    'numpy',
    'scipy',
    'pandas',
    'tensorboard>=2.2.0',
    'toml',
    'torch',
    'torchvision',
    'tqdm',
    'tweetynet>=0.4.2'
]

dev_deps = [
    'twine',
]

test_deps = [
    'vak_test_net',
    'tweetynet>=0.1.1a4',
]

doc_deps = [
    'sphinx',
]

# this is here so that the .travis.yml script can install
# dependencies just for the tests by running
# pip install .[tests]
EXTRAS = {
    'tests': test_deps,
    'dev': dev_deps,
    'doc': doc_deps,
}

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(
            '{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()


METRICS = [
    'Accuracy=vak.metrics.Accuracy',
    'Levenshtein=vak.metrics.Levenshtein',
    'SegmentErrorRate=vak.metrics.SegmentErrorRate',
]

ENTRY_POINTS = {
    'console_scripts': ['vak=vak.__main__:main'],
    'vak.models': [],
    'vak.metrics': METRICS,
}


# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(where="src", exclude=('tests',)),
    package_dir={"": "src"},
    # scripts=['src/bin/vak-cli.py'],
    entry_points=ENTRY_POINTS,
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license=LICENSE,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: BSD License',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



"""sub-package that parses config.toml files and returns config object"""
from . import models
from . import parse
from . import validators

from .learncurve import parse_learncurve_config, LearncurveConfig
from .parse import Config
from .prep import parse_prep_config, PrepConfig
from .predict import parse_predict_config, PredictConfig
from .spect_params import parse_spect_params_config, SpectParamsConfig
from .train import parse_train_config, TrainConfig



from pathlib import Path
from distutils.util import strtobool


def bool_from_str(value):
    if type(value) == bool:
        return value
    elif type(value) == str:
        return bool(strtobool(value))


def comma_separated_list(value):
    if type(value) is list:
        return value
    elif type(value) is str:
        return [element.strip() for element in value.split()]
    else:
        raise TypeError(
            f'unexpected type when converting to comma-separated list: {type(value)}'
        )


def expanded_user_path(value):
    return Path(value).expanduser()


def range_str(range_str, sort=True):
    """Generate range of ints from a formatted string,
    then convert range from int to str

    Examples
    --------
    >>> range_str('1-4,6,9-11')
    ['1','2','3','4','6','9','10','11']

    Takes a range in form of "a-b" and returns
    a list of numbers between a and b inclusive.
    Also accepts comma separated ranges like "a-b,c-d,f"  which will
    return a list with numbers from a to b, c to d, and f.

    Parameters
    ----------
    range_str : str
        of form 'a-b,c', where a hyphen indicates a range
        and a comma separates ranges or single numbers
    sort : bool
        If True, sort output before returning. Default is True.

    Returns
    -------
    list_range : list
        of integer values converted to single-character strings, produced by parsing range_str
    """
    # adapted from
    # http://code.activestate.com/recipes/577279-generate-list-of-numbers-from-hyphenated-and-comma/
    s = "".join(range_str.split())  # removes white space
    list_range = []
    for substr in range_str.split(','):
        subrange = substr.split('-')
        if len(subrange) not in [1, 2]:
            raise SyntaxError("unable to parse range {} in labelset {}."
                              .format(subrange, substr))
        list_range.extend(
            [int(subrange[0])]
        ) if len(subrange) == 1 else list_range.extend(
            range(int(subrange[0]), int(subrange[1]) + 1))

    if sort:
        list_range.sort()

    return [str(list_int) for list_int in list_range]


def labelset_from_toml_value(value):
    """convert value for 'labelset' option from .toml config file into a set

    Parameters
    ----------
    value : str, list
        value assigned to 'labelset' option in a .toml file

    Returns
    -------
    labelset : set
        of single-character strings.

    Notes
    -----
    If value is a str, and it starts with "range:", then everything after range is converted to
    some range of integers, by passing the string to vak.config.converters.range_str,
    and the returned list is converted to a set. E.g. "range: 1-5" becomes {'1', '2', '3', '4', '5'}.
    Other strings that do not start with "range:" are just converted to a set. E.g. "abc" becomes {'a', 'b', 'c'}.

    If value is a list, returns set(value). If all values in list are int,
    they will be converted to strings, e.g. [1, 2, 3, 4, 5] becomes {'1', '2', '3', '4', '5'}.
    If all values are string, any that begin with "range:" will be passed to vak.config.converters.range_str.

    Examples
    --------
    >>> labelset_from_toml_value('1235')
    {'1', '2', '3', '5'}
    >>> labelset_from_toml_value('range: 1-3, 5')
    {'1', '2', '3', '5'}
    >>> labelset_from_toml_value([1, 2, 3])
    {'1', '2', '3'}
    >>> labelset_from_toml_value(['a', 'b', 'c'])
    {'a', 'b', 'c'}
    >>> labelset_from_toml_value(['range: 1-3', 'a-'])
    {'-', '1', '2', '3', 'a'}
    """
    if type(value) is str:
        if value.startswith('range:'):
            value = value.replace('range:', '')
            return set(range_str(value))
        else:
            return set(value)
    elif type(value) is list:
        if all([type(el) == int for el in value]):
            return set([str(el) for el in value])
        elif all([type(el) == str for el in value]):
            labelset = []
            for el in value:
                if el.startswith('range:'):
                    el = el.replace('range:', '')
                    labelset.extend(range_str(el))
                else:
                    labelset.extend(
                        list(el)  # in case element is a string with multiple characters
                    )
            return set(labelset)



from datetime import datetime
from pathlib import Path

from .. import config
from .. import core
from .. import logging


def eval(toml_path):
    """evaluate a trained model with dataset specified in config.toml file.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml(toml_path)

    if cfg.eval is None:
        raise ValueError(
            f'eval called with a config.toml file that does not have a EVAL section: {toml_path}'
        )

    # ---- set up logging ----------------------------------------------------------------------------------------------
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    logger = logging.get_logger(log_dst=cfg.eval.output_dir,
                                caller='eval',
                                timestamp=timenow,
                                logger_name=__name__)
    logger.info('Logging results to {}'.format(cfg.eval.output_dir))

    model_config_map = config.models.map_from_path(toml_path, cfg.eval.models)

    core.eval(cfg.eval.csv_path,
              model_config_map,
              checkpoint_path=cfg.eval.checkpoint_path,
              labelmap_path=cfg.eval.labelmap_path,
              output_dir=cfg.eval.output_dir,
              window_size=cfg.dataloader.window_size,
              num_workers=cfg.eval.num_workers,
              spect_scaler_path=cfg.eval.spect_scaler_path,
              spect_key=cfg.spect_params.spect_key,
              timebins_key=cfg.spect_params.timebins_key,
              device=cfg.eval.device,
              logger=logger)



from pathlib import Path

import attr
from attr.validators import instance_of, optional
import toml
from toml.decoder import TomlDecodeError

from .dataloader import parse_dataloader_config, DataLoaderConfig
from .eval import parse_eval_config, EvalConfig
from .learncurve import parse_learncurve_config, LearncurveConfig
from .predict import parse_predict_config, PredictConfig
from .prep import parse_prep_config, PrepConfig
from .spect_params import parse_spect_params_config, SpectParamsConfig
from .train import parse_train_config, TrainConfig

from .validators import are_sections_valid, are_options_valid


@attr.s
class Config:
    """class to represent config.toml file

    Attributes
    ----------
    prep : vak.config.prep.PrepConfig
        represents [PREP] section of config.toml file
    spect : vak.config.spectrogram.SpectConfig
        represents [SPECTROGRAM] section of config.toml file
    dataloader : vak.config.dataloader.DataLoaderConfig
        represents [DATALOADER] section of config.toml file
    train : vak.config.train.TrainConfig
        represents [TRAIN] section of config.toml file
    eval : vak.config.eval.EvalConfig
        represents [EVAL] section of config.toml file
    predict : vak.config.predict.PredictConfig
        represents [PREDICT] section of config.toml file.
    learncurve : vak.config.learncurve.LearncurveConfig
        represents [LEARNCURVE] section of config.toml file
    """
    spect_params = attr.ib(validator=instance_of(SpectParamsConfig), default=SpectParamsConfig())
    dataloader = attr.ib(validator=instance_of(DataLoaderConfig), default=DataLoaderConfig())

    prep = attr.ib(validator=optional(instance_of(PrepConfig)), default=None)
    train = attr.ib(validator=optional(instance_of(TrainConfig)), default=None)
    eval = attr.ib(validator=optional(instance_of(EvalConfig)), default=None)
    predict = attr.ib(validator=optional(instance_of(PredictConfig)), default=None)
    learncurve = attr.ib(validator=optional(instance_of(LearncurveConfig)), default=None)


SECTION_PARSERS = {
    'SPECT_PARAMS': parse_spect_params_config,
    'DATALOADER': parse_dataloader_config,
    'PREP': parse_prep_config,
    'EVAL': parse_eval_config,
    'TRAIN': parse_train_config,
    'LEARNCURVE': parse_learncurve_config,
    'PREDICT': parse_predict_config,
}


def from_toml(toml_path):
    """parse a TOML configuration file

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format

    Returns
    -------
    config : vak.config.parse.Config
        instance of Config class, whose attributes correspond to
        sections in a config.toml file.
    """
    # check config_path is a file,
    # because if it doesn't ConfigParser will just return an "empty" instance w/out sections or options
    toml_path = Path(toml_path)
    if not toml_path.is_file():
        raise FileNotFoundError(f'path not recognized as a file: {toml_path}')

    try:
        with toml_path.open('r') as fp:
            config_toml = toml.load(fp)
    except TomlDecodeError as e:
        raise Exception(f'Error when parsing .toml config file: {toml_path}') from e

    are_sections_valid(config_toml, toml_path)

    if 'TRAIN' in config_toml:
        if 'LEARNCURVE' in config_toml:
            raise ValueError(
                'a single config.toml file cannot contain both TRAIN and LEARNCURVE sections, '
                'because it is unclear which of those two sections to add paths to when running '
                'the "prep" command to prepare datasets'
            )

        if 'PREP' in config_toml:
            if 'test_set_duration' in config_toml['PREP']:
                raise ValueError(
                    "cannot define 'test_set_duration' option for PREP section when using with vak 'train' command, "
                    "'test_set_duration' is not a valid option for the TRAIN section. "
                    "Were you trying to use the 'learncurve' command instead?"
                )

    config_dict = {}
    for section_name in SECTION_PARSERS.keys():
        if section_name in config_toml:
            are_options_valid(config_toml, section_name, toml_path)
            section_parser = SECTION_PARSERS[section_name]
            config_dict[section_name.lower()] = section_parser(config_toml, toml_path)

    return Config(**config_dict)



from pathlib import Path
import shutil
from datetime import datetime

from .. import config
from .. import core
from .. import logging


def learning_curve(toml_path):
    """generate learning curve, by training models on training sets across a
    range of sizes and then measure accuracy of those models on a test set.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.

    Returns
    -------
    None

    Trains models, saves results in new directory within root_results_dir specified
    in config.toml file, and adds path to that new directory to config.toml file.
    """
    toml_path = Path(toml_path)
    cfg = config.parse.from_toml(toml_path)

    if cfg.learncurve is None:
        raise ValueError(
            f'train called with a config.toml file that does not have a TRAIN section: {toml_path}'
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    timenow = datetime.now().strftime('%y%m%d_%H%M%S')
    results_dirname = f'results_{timenow}'
    if cfg.learncurve.root_results_dir:
        results_path = Path(cfg.learncurve.root_results_dir)
    else:
        results_path = Path('.')
    results_path = results_path.joinpath(results_dirname)
    results_path.mkdir(parents=True)
    # copy config file into results dir now that we've made the dir
    shutil.copy(toml_path, results_path)

    # ---- set up logging ----------------------------------------------------------------------------------------------
    logger = logging.get_logger(log_dst=results_path,
                                caller='train',
                                timestamp=timenow,
                                logger_name=__name__)
    logger.info('Logging results to {}'.format(results_path))

    model_config_map = config.models.map_from_path(toml_path, cfg.learncurve.models)

    core.learning_curve(model_config_map,
                        cfg.learncurve.train_set_durs,
                        cfg.learncurve.num_replicates,
                        cfg.learncurve.csv_path,
                        cfg.prep.labelset,
                        cfg.dataloader.window_size,
                        cfg.learncurve.batch_size,
                        cfg.learncurve.num_epochs,
                        cfg.learncurve.num_workers,
                        results_path=results_path,
                        spect_key=cfg.spect_params.spect_key,
                        timebins_key=cfg.spect_params.timebins_key,
                        normalize_spectrograms=cfg.learncurve.normalize_spectrograms,
                        shuffle=cfg.learncurve.shuffle,
                        val_step=cfg.learncurve.val_step,
                        ckpt_step=cfg.learncurve.ckpt_step,
                        patience=cfg.learncurve.patience,
                        device=cfg.learncurve.device,
                        logger=logger,
                        )
